using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Phidget22;
using System.Threading;
using Phidget22.ExampleUtils;

namespace CompassCalibrator {
	public partial class Form1 : Form {
		static int samplesPerSample = 5;
		Thread calibrateThread;

		CommandLineOpen open;
		formCleaner cleaner;

		bool calibrated = false;

		Spatial spatial;
		bool sampling = false;
		int sampleCounter = 0;

		Calibration cal;

		List<compassDataPoint> compassSamples = new List<compassDataPoint>();

		Point origin = new Point(0, 0);
		Thread drawThread;
		Pipeline p;
		int zoom = 6000;
		Vector3 rots;

		private ErrorEventBox errorBox;

		#region Init/Deinit
		public Form1() {
			open = new CommandLineOpen(this);
			cleaner = new formCleaner(this);
			InitializeComponent();
		}
		public Form1(String[] commandLine) {
			open = new CommandLineOpen(this);
			cleaner = new formCleaner(this);
			open.commandLine = commandLine;
			InitializeComponent();
		}

		private void Form1_Load(object sender, EventArgs e) {

			errorBox = new ErrorEventBox(this);

			p = new Pipeline();

			//add some initial points
			p.addPoint(new Vector3(0, 0, 0), Brushes.Black);
			p.addPoint(new Vector3(1, 0, 0), Brushes.Black);
			p.addPoint(new Vector3(0, 1, 0), Brushes.Black);
			p.addPoint(new Vector3(0, 0, -1), Brushes.Black);
			p.addPoint(new Vector3(0, 0, 0), Brushes.Black);
			p.addPoint(new Vector3(0, 0, 0), Brushes.Black);

			p.indexBuffer.AddRange(new int[] { 0, 1 });
			p.indexBuffer.AddRange(new int[] { 0, 2 });
			p.indexBuffer.AddRange(new int[] { 0, 3 });
			p.indexBuffer.AddRange(new int[] { 0, 4 });
			p.indexBuffer.AddRange(new int[] { 0, 5 });

			p.lineColors.AddRange(new Pen[] { Pens.Black, Pens.Black, Pens.Black, Pens.Red, Pens.Green });

			//rotation speed
			rots = new Vector3(0, 0.01, 0);
			Vector3 initialRots = new Vector3(0, 0, 0);

			Matrix3x3 rotMatrix = new Matrix3x3();
			rotMatrix.matrix[0, 0] = 1; rotMatrix.matrix[1, 0] = 0; rotMatrix.matrix[2, 0] = 0;
			rotMatrix.matrix[0, 1] = 0; rotMatrix.matrix[1, 1] = 1; rotMatrix.matrix[2, 1] = 0;
			rotMatrix.matrix[0, 2] = 0; rotMatrix.matrix[1, 2] = 0; rotMatrix.matrix[2, 2] = 1;
			rotMatrix = Math3D.nextRotMatrix2(rotMatrix, initialRots);
			p.rotMatrix = rotMatrix;

			origin = new Point(pictureBox1.Width / 2, pictureBox1.Height / 2);
			Math3D.RotatePoints(p.rotMatrix, p.vertexBuffer, p.originalVertices);
			pictureBox1.Image = p.Draw(origin, zoom);
			trackBar4.Value = zoom;

			spatial = new Spatial();
			spatial.Attach += Spatial_Attach;
			spatial.Detach += Spatial_Detach;
			spatial.Error += Spatial_Error;
			spatial.SpatialData += Spatial_SpatialData;

			commandLineData phidgetParameters = open.parseCmdLine(); //get command line parameters

			try { //set all the values grabbed from command line.  these values have defaults that are set in ExampleUtils.cs, you can check there to see them.
				spatial.Channel = phidgetParameters.Channel; //selects the channel on the device to open
				spatial.DeviceSerialNumber = phidgetParameters.SerialNumber; //selects the device or hub to open
				spatial.HubPort = phidgetParameters.HubPort; //selects the port on the hub to open
				spatial.IsHubPortDevice = phidgetParameters.isHubPortDevice; //is the device a port on a VINT hub?

				if (phidgetParameters.isRemote) //are we trying to open a remote device?
				{
					spatial.IsRemote = true;
					Net.EnableServerDiscovery(ServerType.Device); //turn on network scan
					if (phidgetParameters.Password != null && phidgetParameters.ServerName != null)
						Net.SetServerPassword(phidgetParameters.ServerName, phidgetParameters.Password); //set the password if there is one
				} else
					spatial.IsLocal = true;

				spatial.Open(); //open the device specified by the above parameters
			} catch (PhidgetException ex) { errorBox.addMessage("Error opening device: " + ex.Message); }
		}

		private void Spatial_Error(object sender, Phidget22.Events.ErrorEventArgs e) {
			errorBox.addMessage(e.Description);
		}

		//When the application is being terminated, close the Phidget.
		private void Form1_FormClosing(object sender, FormClosingEventArgs e) {
			if (drawThread != null)
				if (drawThread.IsAlive)
					drawThread.Abort();

			spatial.Attach -= Spatial_Attach;
			spatial.Detach -= Spatial_Detach;
			spatial.SpatialData -= Spatial_SpatialData;

			spatial.Close();
		}
		#endregion

		#region Data Sampling
		private void Spatial_Attach(object sender, Phidget22.Events.AttachEventArgs e) {
			Spatial attached = (Spatial)sender;
			phidgetInfoBox1.FillPhidgetInfo((Phidget)sender);

			addMessage("Press start to begin calibration." + Environment.NewLine + "For more accuracy, you can enter the local field strength, which can be calculated for your location here: https://www.ngdc.noaa.gov/geomag-web/#igrfwmm");
			sampleDoneButton.Enabled = true;
			spatial.DataInterval = spatial.MinDataInterval;

			//Do this in a seperate thread so the drawing doens't block the data event and cause errors.
			drawThread = new Thread(new ThreadStart(drawThreadFunc));
			drawThread.Start();
		}

		private void Spatial_Detach(object sender, Phidget22.Events.DetachEventArgs e) {
			Spatial detached = (Spatial)sender;
			phidgetInfoBox1.Clear();
			cleaner.clean();
			sampleDoneButton.Enabled = false;

			pictureBox1.Image = null;
			messagesTextBox.Clear();

			if (drawThread != null)
				if (drawThread.IsAlive)
					drawThread.Abort();
		}

		private void Spatial_SpatialData(object sender, Phidget22.Events.SpatialSpatialDataEventArgs e) {
			//This displays a vector representing the measured field in red, and after calibration the calibrated field in green
			if (e.MagneticField[0] == (double)1e300 || e.MagneticField[1] == (double)1e300 || e.MagneticField[2] == (double)1e300)
				return;
			lock (p) {
				if (twoAxisButton.Checked) {
					p.originalVertices[4] = new Vector3(
						e.MagneticField[0],
						e.MagneticField[1],
						-e.MagneticField[2]);
					if (calibrated) {
						Vector3 rawData = new Vector3(e.MagneticField[0], e.MagneticField[1], e.MagneticField[2]);
						Vector3 calibratedData = Matrix3x3.Multiply(Vector3.Subtract(rawData, cal.offset), cal.gainAndTransform);

						//apply local mag field - calibrate returns a unit (mag 1) field
						calibratedData = Vector3.Multiply(double.Parse(localFieldTextBox.Text), calibratedData);

						p.originalVertices[5] = new Vector3(
							calibratedData.X,
							calibratedData.Y,
							-calibratedData.Z);
					}
				} else {
					p.originalVertices[4] = new Vector3(
						e.MagneticField[0],
						e.MagneticField[2],
						-e.MagneticField[1]);
					if (calibrated) {
						Vector3 rawData = new Vector3(e.MagneticField[0], e.MagneticField[1], e.MagneticField[2]);
						Vector3 calibratedData = Matrix3x3.Multiply(Vector3.Subtract(rawData, cal.offset), cal.gainAndTransform);

						//apply local mag field - calibrate returns a unit (mag 1) field
						calibratedData = Vector3.Multiply(double.Parse(localFieldTextBox.Text), calibratedData);

						p.originalVertices[5] = new Vector3(
							calibratedData.X,
							calibratedData.Z,
							-calibratedData.Y);
					}
				}
			}

			//This samples magnetic field data
			if (sampling) {
				lock (compassSamples) {
					if (sampling) {
						sampleCounter++;
						if (sampleCounter >= samplesPerSample) {
							sampleCounter = 0;
							compassSamples.Add(
								new compassDataPoint(
								e.MagneticField[0],
								e.MagneticField[1],
								e.MagneticField[2]
								)
							);

							//This draws the magnetic vectors as dots
							lock (p) {
								if (twoAxisButton.Checked) {
									p.addPoint(new Vector3(
										e.MagneticField[0],
										e.MagneticField[1],
										-e.MagneticField[2]), Brushes.Red);
								} else {
									p.addPoint(new Vector3(
										e.MagneticField[0],
										e.MagneticField[2],
										-e.MagneticField[1]), Brushes.Red);
								}
							}
						}
					}
				}
			}
		}
		#endregion

		#region Calibration
		//Runs the calibration algorithm
		void calibrateThreadFunction() {
			addMessage("Starting Calibration with " + compassSamples.Count.ToString() + " samples.");
			cal = new Calibration(this);
			bool calSuccess = false;
			if (threeAxisButton.Checked)
				calSuccess = cal.calibrate3(compassSamples);
			else
				calSuccess = cal.calibrate2(compassSamples);
			if (calSuccess) {
				//Calibration was successful - redraw all the samples in green.
				addMessage("Calibration successful.");
				if (localFieldTextBox.Text == "")
					setLocalField(cal.estimatedLocalField.ToString("F5"));
				for (int i = 0; i < compassSamples.Count; i++) {
					//apply calibration
					Vector3 rawData = new Vector3(compassSamples[i].data[0], compassSamples[i].data[1], compassSamples[i].data[2]);
					Vector3 calibratedData = Matrix3x3.Multiply(Vector3.Subtract(rawData, cal.offset), cal.gainAndTransform);

					//apply local mag field - calibrate returns a unit (mag 1) field
					calibratedData = Vector3.Multiply(double.Parse(localFieldTextBox.Text), calibratedData);

					lock (p) {
						if (twoAxisButton.Checked) {
							p.addPoint(new Vector3(
								calibratedData.X,
								calibratedData.Y,
								-calibratedData.Z), Brushes.Green);
						} else {
							p.addPoint(new Vector3(
								calibratedData.X,
								calibratedData.Z,
								-calibratedData.Y), Brushes.Green);
						}
					}
				}
				try {
					switch (spatial.DeviceID) {

					case DeviceID.PN_1056:
						addMessage("These calibration values must be set in software EVERY TIME the spatial is opened/attached, " +
						"as they are maintained in the PC-side library code. This is best done in the attach handler.");
						break;

					default:
						//Actually set the calibration
						spatial.SetMagnetometerCorrectionParameters(cal.estimatedLocalField, cal.offset.X, cal.offset.Y, cal.offset.Z,
							cal.gainAndTransform.matrix[0, 0], cal.gainAndTransform.matrix[1, 1], cal.gainAndTransform.matrix[2, 2],
							cal.gainAndTransform.matrix[0, 1], cal.gainAndTransform.matrix[0, 2], cal.gainAndTransform.matrix[1, 0],
							cal.gainAndTransform.matrix[1, 2], cal.gainAndTransform.matrix[2, 0], cal.gainAndTransform.matrix[2, 1]);
						spatial.SaveMagnetometerCorrectionParameters();

						addMessage("Calibration values have been written to firmware on your spatial. " +
							"These values will be maintained from now on, across power cycles, until explicitely reset or changed.");
						break;
					}

					calibrated = true;
				} catch (PhidgetException ex) { errorBox.addMessage("Failed to write calibration values: " + ex.Message); }
			} else {
				//failed
				addMessage("Calibration failed, try again.");
			}

			setButtonEnabled(sampleDoneButton, true, "Start");
		}

		private void sampleDoneButton_Click(object sender, EventArgs e) {
			if (sampling) {
				sampling = false;
				lock (compassSamples) {
					sampleDoneButton.Text = "wait...";
					sampleDoneButton.Enabled = false;
					//start the calibration thread
					calibrateThread = new Thread(new ThreadStart(calibrateThreadFunction));
					calibrateThread.Start();
				}
			} else {
				sampling = false;
				lock (compassSamples) {
					calibrated = false;
					//clear graph
					resetGraph();
					try {
						switch (spatial.DeviceID) {
						case DeviceID.PN_1056:
							break;

						default:
							//Reset any previous calibration values
							addMessage("Clearing any previous calibration data...");
							spatial.ResetMagnetometerCorrectionParameters();
							Thread.Sleep(100);
							break;
						}
					} catch (PhidgetException ex) {
						errorBox.addMessage("Failed to reset calibration values: " + ex.Message);
						return;
					}

					compassSamples = new List<compassDataPoint>();
					addMessage("Starting sampling... start moving device now.");
					if (threeAxisButton.Checked)
						addMessage("Try to draw out as much of a sphere as possible, then press stop");
					else
						addMessage("Rotate the board about the z-axis. This should draw out a circle. Press stop when finished.");
					sampling = true;
					sampleDoneButton.Text = "Stop";
				}
			}
		}

		private void twoAxisButton_CheckedChanged(object sender, EventArgs e) {
			lock (p) {
				if (twoAxisButton.Checked) {
					threeAxisButton.Checked = false;
					rots = new Vector3(0, 0, 0);
				} else {
					threeAxisButton.Checked = true;
					rots = new Vector3(0, 0.01, 0);
				}
				resetGraph();
			}
		}

		private void threeAxisButton_CheckedChanged(object sender, EventArgs e) {
			lock (p) {
				if (threeAxisButton.Checked) {
					twoAxisButton.Checked = false;
					rots = new Vector3(0, 0.01, 0);
				} else {
					twoAxisButton.Checked = true;
					rots = new Vector3(0, 0, 0);
				}
				resetGraph();
			}
		}
		#endregion

		#region Point Graph
		//Draws the rotating graph of data points
		void drawThreadFunc() {
			while (true) {
				try {
					Bitmap g;
					lock (p) {
						p.rotMatrix = Math3D.nextRotMatrix2(p.rotMatrix, rots);
						Math3D.RotatePoints(p.rotMatrix, p.vertexBuffer, p.originalVertices);
						g = p.Draw(origin, zoom);
					}
					setPicture(pictureBox1, g);
					Thread.Sleep(10);
				} catch (ThreadAbortException) {
					Console.WriteLine("drawThread Aborted");
					return;
				} catch (Exception e) {
					Console.WriteLine(e.Message);
				}
			}
		}

		private void resetGraph() {
			lock (p) {
				Matrix3x3 rotMatrix = new Matrix3x3();
				rotMatrix.matrix[0, 0] = 1; rotMatrix.matrix[1, 0] = 0; rotMatrix.matrix[2, 0] = 0;
				rotMatrix.matrix[0, 1] = 0; rotMatrix.matrix[1, 1] = 1; rotMatrix.matrix[2, 1] = 0;
				rotMatrix.matrix[0, 2] = 0; rotMatrix.matrix[1, 2] = 0; rotMatrix.matrix[2, 2] = 1;
				p.rotMatrix = rotMatrix;

				p.originalVertices.Clear();
				p.vertexBuffer.Clear();
				p.pointColors.Clear();

				//add some initial points
				p.addPoint(new Vector3(0, 0, 0), Brushes.Black);
				p.addPoint(new Vector3(1, 0, 0), Brushes.Black);
				p.addPoint(new Vector3(0, 1, 0), Brushes.Black);
				p.addPoint(new Vector3(0, 0, -1), Brushes.Black);
				p.addPoint(new Vector3(0, 0, 0), Brushes.Black);
				p.addPoint(new Vector3(0, 0, 0), Brushes.Black);
			}
		}

		private void pictureBox1_Resize(object sender, EventArgs e) {
			origin = new Point(pictureBox1.Width / 2, pictureBox1.Height / 2);
		}

		private void trackBar4_Scroll(object sender, EventArgs e) {
			zoom = trackBar4.Value;
		}

		private void rotationTrack_Scroll(object sender, EventArgs e) {
			if (threeAxisButton.Checked)
				rots.Y = (rotationTrack.Value / 10000.0);
			else
				rots.Z = (rotationTrack.Value / 10000.0);
		}
		#endregion

		#region Invokers
		public delegate void StringDelegate(String str);
		public void addMessage(String message) {
			if (messagesTextBox.InvokeRequired)
				try { messagesTextBox.Invoke(new StringDelegate(addMessage), new Object[] { message }); } catch { } else {
				messagesTextBox.SelectionColor = Color.Black;
				messagesTextBox.AppendText(message);
				if (messagesTextBox.Text.Length > 2)
					messagesTextBox.Select(messagesTextBox.Text.Length - 1, 1);
				messagesTextBox.ScrollToCaret();
				messagesTextBox.AppendText(Environment.NewLine);
			}
		}
		public void setLocalField(String txt) {
			if (localFieldTextBox.InvokeRequired)
				try { localFieldTextBox.Invoke(new StringDelegate(setLocalField), new Object[] { txt }); } catch { } else {
				localFieldTextBox.Text = txt;
			}
		}
		public delegate void SetButtonEnabledDelegate(Button button, Boolean enabled, String text);
		public void setButtonEnabled(Button button, Boolean enabled, String text) {
			if (button.InvokeRequired) {
				try {
					button.Invoke(new SetButtonEnabledDelegate(setButtonEnabled), new Object[] { button, enabled, text });
				} catch { }
			} else {
				button.Enabled = enabled;
				button.Text = text;
			}
		}
		public delegate void PictureBoxDelegate(PictureBox box, Image pic);
		public void setPicture(PictureBox box, Image pic) {
			if (box.InvokeRequired) {
				try { box.Invoke(new PictureBoxDelegate(setPicture), new Object[] { box, pic }); } catch { }
			} else
				box.Image = pic;
		}
		#endregion
	}
}